import argh
import os
import json
import re
import subprocess
import tempfile
import sys
import shutil
import contextlib
import logging
import random
import time
import gzip
import io
import tarfile
import threading
import fnmatch
import socket
import ssl
from datetime import datetime
from hashlib import sha1
from pathlib import Path
from functools import partial
from itertools import cycle
from typing import Dict, Any, List, NamedTuple
from urllib.request import urlopen

from cr8.misc import parse_version, init_logging
from cr8.engine import DotDict
from cr8.exceptions import ArgumentError

log = logging.getLogger(__name__)

NO_SSL_VERIFY_CTX = ssl._create_unverified_context()
RELEASE_URL = 'https://cdn.crate.io/downloads/releases/crate-{version}.tar.gz'
VERSION_RE = re.compile(r'^(\d+\.\d+\.\d+)$')
DYNAMIC_VERSION_RE = re.compile(r'^((\d+|x)\.(\d+|x)\.(\d+|x))$')
BRANCH_VERSION_RE = re.compile(r'^((\d+)\.(\d+))$')
FOLDER_VERSION_RE = re.compile(r'crate-(\d+\.\d+\.\d+)')
REPO_URL = 'https://github.com/crate/crate.git'

DEFAULT_SETTINGS = {
    'discovery.initial_state_timeout': 0,
    'network.host': '127.0.0.1',
    'udc.enabled': False
}


def _format_cmd_option_legacy(k, v):
    return '-Des.{0}={1}'.format(k, v)


def _format_cmd_option(k, v):
    if isinstance(v, bool):
        return '-C{0}={1}'.format(k, str(v).lower())
    return '-C{0}={1}'.format(k, v)


def _extract_version(crate_dir) -> tuple:
    m = FOLDER_VERSION_RE.findall(crate_dir)
    if m:
        return parse_version(m[0])
    return (1, 0, 0)


class OutputMonitor:

    def __init__(self):
        self.consumers = []

    def _consume(self, proc):
        try:
            for line in proc.stdout:
                for consumer in self.consumers:
                    consumer.send(line)
        except Exception:
            if proc.returncode is not None:
                return
            raise

    def start(self, proc):
        out_thread = threading.Thread(target=self._consume, args=(proc,))
        out_thread.daemon = True
        out_thread.start()


class Timeout:
    def __init__(self, timeout, sleep=0.1):
        self.start_time = time.time()
        self.sleep = sleep
        self._first_ok = True
        self.timeout = timeout

        def timeout_expired():
            if self._first_ok:
                self._first_ok = False
                return False
            now = time.time()
            if (now - self.start_time) > self.timeout:
                return True
            if self.sleep:
                time.sleep(self.sleep)

        self._timeout_expired = timeout_expired

    def __call__(self):
        if self._timeout_expired():
            raise TimeoutError()
        return True


def wait_until(predicate, timeout=30):
    """Wait until predicate returns a truthy value or the timeout is reached.

    >>> wait_until(lambda: True, timeout=10)
    """
    not_expired = Timeout(timeout)
    while not_expired():
        r = predicate()
        if r:
            break


def _is_up(host: str, port: int):
    try:
        conn = _create_connection(host, port)
        conn.close()
        return True
    except (socket.gaierror, ConnectionRefusedError):
        return False


def _create_connection(host: str, port: int):
    if host[0] == '[' and host[-1] == ']':
        host = host[1:-1]
    return socket.create_connection((host, port))


def _has_ssl(host: str, port: int):
    try:
        with ssl.wrap_socket(_create_connection(host, port)) as s:
            s.close()
            return True
    except (socket.gaierror, ssl.SSLError):
        return False


def cluster_state_200(url):
    try:
        with urlopen(url, context=NO_SSL_VERIFY_CTX) as r:
            p = json.loads(r.read().decode('utf-8'))
            return int(p['status']) == 200
    except Exception as e:
        log.debug(e)
        return False


def _get_settings(settings=None):
    s = DEFAULT_SETTINGS.copy()
    if settings:
        s.update(settings)
    return s


def _try_print_log(logfile):
    try:
        with open(logfile) as f:
            for line in f:
                log.error(line)
    except Exception:
        pass


def _ensure_running(proc):
    result = proc.poll()
    if result:
        raise SystemError('Process exited: ' + str(result))
    return True


class CrateNode(contextlib.ExitStack):
    """Class that allows starting and stopping a Crate process

    This is similar to the ``CrateLayer`` in ``crate.testing.layer``.
    But additionaly it supports setting environment variables and it can infer
    the port to which Crate binds by sniffing Crate's stdout.

    Attributes:
        http_url: The HTTP URL of the Crate process.
            Only available after ``start()`` has been called.
        process: The subprocess. Only available after ``start()`` has been called.
    """

    def __init__(self,
                 crate_dir: str,
                 env: Dict[str, Any] = None,
                 settings: Dict[str, Any] = None,
                 keep_data: bool = False) -> None:
        """Create a CrateNode

        Args:
            crate_dir: Path to the extracted Crate tarball
            env: Environment variables with which the Crate process will be
                started.
            settings: Additional Crate settings.
        """
        super().__init__()
        self.crate_dir = crate_dir
        version = _extract_version(crate_dir)
        self.env = env or {}
        self.env.setdefault('JAVA_HOME',
                            os.environ.get('JAVA_HOME', ''))
        self.env.setdefault('LANG',
                            os.environ.get('LANG', os.environ.get('LC_ALL')))
        if not self.env['LANG']:
            raise SystemError('Your locale are not configured correctly. '
                              'Please set LANG or alternatively LC_ALL.')
        self.monitor = OutputMonitor()
        self.process = None  # type: subprocess.Popen
        self.http_url = None  # type: str
        self.http_host = None  # type: str
        start_script = 'crate.bat' if sys.platform == 'win32' else 'crate'

        settings = _get_settings(settings)
        if version < (1, 1, 0):
            settings.setdefault('discovery.zen.ping.multicast.enabled', False)
        self.data_path = settings.get('path.data') or tempfile.mkdtemp()
        self.logs_path = settings.get('path.logs') or os.path.join(crate_dir, 'logs')
        self.cluster_name = settings.get('cluster.name') or 'cr8'
        self.keep_data = keep_data
        settings['path.data'] = self.data_path
        settings['cluster.name'] = self.cluster_name
        if version < (1, 0, 0):
            _format_option = _format_cmd_option_legacy
        else:
            _format_option = _format_cmd_option
        args = [_format_option(k, v) for k, v in settings.items()]
        self.cmd = [
            os.path.join(crate_dir, 'bin', start_script)] + args

    def start(self):
        """Start the process.

        This will block until the Crate cluster is ready to process requests.
        """
        log.info('Starting Crate process')
        self.process = proc = self.enter_context(subprocess.Popen(
            self.cmd,
            stdin=subprocess.DEVNULL,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            env=self.env,
            universal_newlines=True
        ))

        msg = ('CrateDB launching:\n'
               '    PID: %s\n'
               '    Logs: %s\n'
               '    Data: %s')
        if not self.keep_data:
            msg += ' (removed on stop)\n'

        logfile = os.path.join(self.logs_path, self.cluster_name + '.log')
        log.info(
            msg,
            proc.pid,
            logfile,
            self.data_path
        )
        self.addresses = DotDict({})
        self.monitor.consumers.append(AddrConsumer(self._set_addr))
        self.monitor.start(proc)

        line_buf = LineBuffer()
        self.monitor.consumers.append(line_buf)
        spinner = cycle(['/', '-', '\\', '|'])

        def show_spinner():
            if sys.stdout.isatty():
                print(next(spinner), end='\r')
            return True
        try:
            wait_until(
                lambda: show_spinner() and _ensure_running(proc) and self.http_host,
                timeout=60
            )
            host = self.addresses.http.host
            port = self.addresses.http.port
            wait_until(
                lambda: _ensure_running(proc) and _is_up(host, port),
                timeout=30
            )
            if _has_ssl(host, port):
                self.http_url = self.http_url.replace('http://', 'https://')
            wait_until(
                lambda: show_spinner() and cluster_state_200(self.http_url),
                timeout=30
            )
        except (SystemError, TimeoutError):
            if not line_buf.lines:
                _try_print_log(logfile)
            else:
                for line in line_buf.lines:
                    log.error(line)
            raise SystemExit("Exiting because CrateDB didn't start correctly")
        else:
            self.monitor.consumers.remove(line_buf)
        log.info('Cluster ready to process requests')

    def _set_addr(self, protocol, addr):
        log.info('{0:10}: {1}'.format(protocol.capitalize(), addr))
        host, port = addr.rsplit(':', 1)
        port = int(port)
        self.addresses[protocol] = Address(host, port)
        if protocol == 'http':
            self.http_host = addr
            self.http_url = 'http://' + addr

    def stop(self):
        if self.process:
            self.process.terminate()
            self.process.communicate(timeout=10)
        self.addresses = DotDict({})
        self.http_host = None
        self.http_url = None
        if not self.keep_data:
            path = self.data_path.split(',')
            for p in path:
                shutil.rmtree(p)

    def __enter__(self):
        return self

    def __exit__(self, *ex):
        self.stop()


class LineBuffer:

    def __init__(self):
        self.lines = []

    def send(self, line):
        self.lines.append(line.strip())


class Address(NamedTuple):
    host: str
    port: int


class AddrConsumer:

    ADDRESS_RE = re.compile(
        r'.*\[(?P<protocol>http|i.c.p.h.CrateNettyHttpServerTransport|o.e.h.HttpServer|psql|transport|o.e.t.TransportService)\s*\] \[.*\] .*'
        r'publish_address {'
        r'(?:(inet\[[A-Za-z-\.]*/)|([A-Za-z\.]*/))?'
        r'?(?P<addr>\[?[\d\.:]+\]?:?\d+)'
        r'(?:\])?'
        r'}'
    )
    PROTOCOL_MAP = {
        'i.c.p.h.CrateNettyHttpServerTransport': 'http',
        'o.e.h.HttpServer': 'http',
        'o.e.t.TransportService': 'transport'
    }

    def __init__(self, on_addr):
        self.on_addr = on_addr

    @staticmethod
    def _parse(line):
        """ Parse protocol and bound address from log message

        >>> AddrConsumer._parse('NONE')
        (None, None)

        >>> AddrConsumer._parse('[INFO ][i.c.p.h.CrateNettyHttpServerTransport] [Widderstein] publish_address {127.0.0.1:4200}, bound_addresses {[fe80::1]:4200}, {[::1]:4200}, {127.0.0.1:4200}')
        ('http', '127.0.0.1:4200')

        >>> AddrConsumer._parse('[INFO ][o.e.t.TransportService   ] [Widderstein] publish_address {127.0.0.1:4300}, bound_addresses {[fe80::1]:4300}, {[::1]:4300}, {127.0.0.1:4300}')
        ('transport', '127.0.0.1:4300')

        >>> AddrConsumer._parse('[INFO ][psql                     ] [Widderstein] publish_address {127.0.0.1:5432}, bound_addresses {127.0.0.1:5432}')
        ('psql', '127.0.0.1:5432')
        """
        m = AddrConsumer.ADDRESS_RE.match(line)
        if not m:
            return None, None
        protocol = m.group('protocol')
        protocol = AddrConsumer.PROTOCOL_MAP.get(protocol, protocol)
        return protocol, m.group('addr')

    def send(self, line):
        protocol, addr = AddrConsumer._parse(line)
        if protocol:
            self.on_addr(protocol, addr)


def _openuri(uri):
    if os.path.isfile(uri):
        return open(uri, 'rb')
    return io.BytesIO(urlopen(uri).read())


def _can_use_cache(uri, crate_dir):
    if not os.path.exists(crate_dir):
        return False
    os.utime(crate_dir)  # update mtime to avoid removal
    if os.path.isfile(uri):
        with _openuri(uri) as f:
            checksum = sha1(f.read()).hexdigest()
            return os.path.exists(os.path.join(crate_dir, checksum))
    # Always enable use of the cache if the source is not local
    return True


def _download_and_extract(uri, crate_root):
    filename = os.path.basename(uri)
    crate_folder_name = re.sub(r'\.tar(\.gz)?$', '', filename)
    crate_dir = os.path.join(crate_root, crate_folder_name)

    if _can_use_cache(uri, crate_dir):
        log.info('Skipping download, tarball alrady extracted at %s', crate_dir)
        return crate_dir
    elif os.path.exists(crate_dir):
        shutil.rmtree(crate_dir, ignore_errors=True)
    log.info('Downloading %s and extracting to %s', uri, crate_root)
    with _openuri(uri) as tmpfile:
        with tarfile.open(fileobj=tmpfile) as t:
            t.extractall(crate_root)
        tmpfile.seek(0)
        checksum = sha1(tmpfile.read()).hexdigest()
        with open(os.path.join(crate_dir, checksum), 'a'):
            os.utime(os.path.join(crate_dir, checksum))
    return crate_dir


def _from_versions_json(key):
    def retrieve():
        with urlopen('https://crate.io/releases.json') as r:
            if r.headers.get('Content-Encoding') == 'gzip':
                with gzip.open(r, 'rt') as r:
                    versions = json.loads(r.read())
            else:
                versions = json.loads(r.read().decode('utf-8'))
        return versions[key]['downloads']['tar.gz']['url']
    return retrieve


RELEASE_RE = re.compile(r'.*>(?P<filename>crate-(?P<version>\d+\.\d+\.\d+)\.tar\.gz)<.*')


def _retrieve_crate_versions():
    base_uri = 'https://cdn.crate.io/downloads/releases/'
    with urlopen(base_uri) as r:
        lines = (line.decode('utf-8') for line in r)
        for line in lines:
            m = RELEASE_RE.match(line)
            if m:
                yield m.group('version')


def _find_matching_version(versions, version_pattern):
    """
    Return the first matching version

    >>> _find_matching_version(['1.1.4', '1.0.12', '1.0.5'], '1.0.x')
    '1.0.12'

    >>> _find_matching_version(['1.1.4', '1.0.6', '1.0.5'], '2.x.x')
    """
    pattern = fnmatch.translate(version_pattern.replace('x', '*'))
    return next((v for v in versions if re.match(pattern, v)), None)


_version_lookups = {
    'latest': _from_versions_json('stable'),
    'latest-stable': _from_versions_json('stable'),
    'latest-testing': _from_versions_json('testing'),
    'latest-nightly': _from_versions_json('nightly')
}


def _lookup_uri(version):
    if version in _version_lookups:
        version = _version_lookups[version]()
    m = VERSION_RE.match(version)
    if m:
        return RELEASE_URL.format(version=m.group(0))
    m = DYNAMIC_VERSION_RE.match(version)
    if m:
        versions = sorted(map(parse_version, list(_retrieve_crate_versions())))
        versions = ['.'.join(map(str, v)) for v in versions[::-1]]
        release = _find_matching_version(versions, m.group(0))
        if release:
            return RELEASE_URL.format(version=release)
    return version


def _is_project_repo(src_repo):
    return (os.path.isdir(src_repo) and
            os.path.exists(os.path.join(src_repo, '.git')) and
            os.path.exists(os.path.join(src_repo, 'gradlew')))


def _build_from_src(src_repo):
    run = partial(subprocess.run, cwd=src_repo, check=True)
    run(['git', 'clean', '-xdff'])
    src_repo = Path(src_repo)
    if os.path.exists(src_repo / 'es' / 'upstream'):
        run(['git', 'submodule', 'update', '--init', '--', 'es/upstream'])
    run(['./gradlew', '--no-daemon', 'clean', 'distTar'])
    distributions = Path(src_repo) / 'app' / 'build' / 'distributions'
    tarball = next(distributions.glob('crate-*.tar.gz'))
    with tarfile.open(tarball) as t:
        t.extractall(tarball.parent)
    # remove two suffixes ('.tar, '.gz') to get the folder name
    return str(tarball.with_suffix('').with_suffix(''))


def _build_from_release_branch(branch, crate_root):
    crates = Path(crate_root)
    src_repo = crates / 'sources'
    run_in_repo = partial(subprocess.run, cwd=src_repo, check=True)
    if not src_repo.exists() or not (src_repo / '.git').exists():
        clone = ['git', 'clone', REPO_URL, 'sources']
        subprocess.run(clone, cwd=crate_root, check=True)
    else:
        run_in_repo(['git', 'fetch'])
    run_in_repo(['git', 'checkout', branch])
    run_in_repo(['git', 'pull', 'origin', branch])
    return _build_from_src(str(src_repo))


def _remove_old_crates(path):
    now = time.time()
    s7days_ago = now - (7 * 24 * 60 * 60)
    with contextlib.suppress(FileNotFoundError):
        old_unused_dirs = (e for e in os.scandir(path)
                           if e.is_dir() and e.stat().st_mtime < s7days_ago)
        for e in old_unused_dirs:
            last_use = datetime.fromtimestamp(e.stat().st_mtime)
            msg = f'Removing from cache: {e.name} (last use: {last_use:%Y-%m-%d %H:%M})'
            print(msg, file=sys.stderr)
            shutil.rmtree(e.path)


def _crates_cache() -> str:
    """ Return the path to the crates cache folder """
    return os.environ.get(
        'XDG_CACHE_HOME',
        os.path.join(os.path.expanduser('~'), '.cache', 'cr8', 'crates'))


def get_crate(version, crate_root=None):
    """Retrieve a Crate tarball, extract it and return the path.

    Args:
        version: The Crate version to get.
            Can be specified in different ways:

            - A concrete version like '0.55.0'
            - A version including a `x` as wildcards. Like: '1.1.x' or '1.x.x'.
              This will use the latest version that matches.
            - Release branch, like `3.1`
            - An alias: 'latest-stable' or 'latest-testing'
            - A URI pointing to a crate tarball
        crate_root: Where to extract the tarball to.
            If this isn't specified ``$XDG_CACHE_HOME/.cache/cr8/crates``
            will be used.
    """
    if not crate_root:
        crate_root = _crates_cache()
        _remove_old_crates(crate_root)
    if _is_project_repo(version):
        return _build_from_src(version)
    m = BRANCH_VERSION_RE.match(version)
    if m:
        return _build_from_release_branch(m.group(0), crate_root)
    uri = _lookup_uri(version)
    crate_dir = _download_and_extract(uri, crate_root)
    return crate_dir


def _parse_options(options: List[str]) -> Dict[str, str]:
    """ Parse repeatable CLI options

    >>> opts = _parse_options(['cluster.name=foo', 'CRATE_JAVA_OPTS="-Dxy=foo"'])
    >>> print(json.dumps(opts, sort_keys=True))
    {"CRATE_JAVA_OPTS": "\\"-Dxy=foo\\"", "cluster.name": "foo"}
    """
    try:
        return dict(i.split('=', maxsplit=1) for i in options)
    except ValueError:
        raise ArgumentError(
            f'Option must be in format <key>=<value>, got: {options}')


def create_node(version, env=None, setting=None, crate_root=None, keep_data=False):
    init_logging(log)
    settings = {
        'cluster.name': 'cr8-crate-run' + str(random.randrange(1e9))
    }
    crate_dir = get_crate(version, crate_root)
    if setting:
        settings.update(_parse_options(setting))
    if env:
        env = _parse_options(env)
    return CrateNode(
        crate_dir=crate_dir, env=env, settings=settings, keep_data=keep_data)


@argh.arg('version', help='Crate version to run')
@argh.arg('-e', '--env', action='append',
          help='Environment variable. Option can be specified multiple times.')
@argh.arg('-s', '--setting', action='append',
          help='Crate setting. Option can be specified multiple times.')
@argh.arg('--keep-data', help='If this is set the data folder will be kept.')
@argh.wrap_errors([ArgumentError])
def run_crate(version, env=None, setting=None, crate_root=None, keep_data=False):
    """Launch a crate instance.

    Supported version specifications:
        - Concrete version like "0.55.0" or with wildcard: "1.1.x"
        - An alias (one of [latest-nightly, latest-stable, latest-testing])
        - A URI pointing to a CrateDB tarball (in .tar.gz format)
        - A URI pointing to a checked out CrateDB repo directory

    run-crate supports command chaining. To launch a CrateDB node and another
    sub-command use:

        cr8 run-crate <ver> -- timeit -s "select 1" --hosts '{node.http_url}'

    To launch any (blocking) subprocess, prefix the name with '@':

        cr8 run-crate <version> -- @http '{node.http_url}'

    If run-crate is invoked using command chaining it will exit once all
    chained commands finished.

    The postgres host and port are available as {node.addresses.psql.host} and
    {node.addresses.psql.port}
    """
    with create_node(version, env, setting, crate_root, keep_data) as n:
        try:
            n.start()
            n.process.wait()
        except KeyboardInterrupt:
            print('Stopping Crate...')


if __name__ == "__main__":
    argh.dispatch_command(run_crate)



import json
import sys
import contextlib
from functools import partial


_wopen = partial(open, mode='w', encoding='utf-8')
to_jsonstr = json.dumps


def _format_short(stats):
    output = ('Runtime (in ms):\n'
              '    mean:    {mean:.3f} ± {error_margin:.3f}')
    values = dict(
        mean=stats['mean'],
        error_margin=stats.get('error_margin', 0.0),
    )
    if stats['n'] > 1:
        output += (
            '\n'
            '    min/max: {min:.3f} → {max:.3f}\n'
            'Percentile:\n'
            '    50:   {p50:.3f} ± {stdev:.3f} (stdev)\n'
            '    95:   {p95:.3f}\n'
            '    99.9: {p999:.3f}'
        )
        percentiles = stats['percentile']
        values.update(dict(
            max=stats['max'],
            min=stats['min'],
            stdev=stats['stdev'],
            p50=percentiles['50'],
            p95=percentiles['95'],
            p999=percentiles['99_9']
        ))
    return output.format(**values)


def format_stats(stats, output_fmt=None):
    output_fmt = output_fmt or 'text'
    if output_fmt == 'json':
        return to_jsonstr(stats)
    return _format_short(stats)


class Logger(contextlib.ExitStack):

    def __init__(self,
                 output_fmt='text',
                 logfile_info=None,
                 logfile_result=None):
        super().__init__()
        info_output = self._open(logfile_info)
        result_output = self._open(logfile_result)
        self.info = partial(print, file=info_output)
        presult = partial(print, file=result_output)
        if output_fmt == 'json':
            self.result = lambda r: presult(to_jsonstr(r.as_dict()))
        else:
            self.result = lambda r: presult(_format_short(r.runtime_stats))

    def _open(self, filepath):
        if filepath:
            return self.enter_context(_wopen(filepath))
        return sys.stdout



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import argh
import sys
from functools import partial

from .cli import dicts_from_stdin, to_int
from .misc import as_bulk_queries
from cr8 import aio, clients
from .metrics import Stats
from .log import format_stats


def to_insert(table, d):
    """Generate an insert statement using the given table and dictionary.

    Args:
        table (str): table name
        d (dict): dictionary with column names as keys and values as values.
    Returns:
        tuple of statement and arguments

    >>> to_insert('doc.foobar', {'name': 'Marvin'})
    ('insert into doc.foobar ("name") values (?)', ['Marvin'])
    """

    columns = []
    args = []
    for key, val in d.items():
        columns.append('"{}"'.format(key))
        args.append(val)
    stmt = 'insert into {table} ({columns}) values ({params})'.format(
        table=table,
        columns=', '.join(columns),
        params=', '.join(['?'] * len(columns)))
    return (stmt, args)


def print_only(table):
    for d in dicts_from_stdin():
        print(to_insert(table, d))
    print('')
    print('No hosts provided. Nothing inserted')


@argh.arg('--table', help='Target table', required=True)
@argh.arg('-b', '--bulk-size', type=to_int)
@argh.arg('--hosts', help='crate hosts which will be used \
          to execute the insert statement', type=str)
@argh.arg('-c', '--concurrency', type=to_int)
@argh.arg('-of', '--output-fmt', choices=['json', 'text'], default='text')
@argh.wrap_errors([KeyboardInterrupt, BrokenPipeError] + clients.client_errors)
def insert_json(table=None,
                bulk_size=1000,
                concurrency=25,
                hosts=None,
                output_fmt=None):
    """Insert JSON lines fed into stdin into a Crate cluster.

    If no hosts are specified the statements will be printed.

    Args:
        table: Target table name.
        bulk_size: Bulk size of the insert statements.
        concurrency: Number of operations to run concurrently.
        hosts: hostname:port pairs of the Crate nodes
    """
    if not hosts:
        return print_only(table)

    queries = (to_insert(table, d) for d in dicts_from_stdin())
    bulk_queries = as_bulk_queries(queries, bulk_size)
    print('Executing inserts: bulk_size={} concurrency={}'.format(
        bulk_size, concurrency), file=sys.stderr)

    stats = Stats()
    with clients.client(hosts, concurrency=concurrency) as client:
        f = partial(aio.measure, stats, client.execute_many)
        try:
            aio.run_many(f, bulk_queries, concurrency)
        except clients.SqlException as e:
            raise SystemExit(str(e))
    try:
        print(format_stats(stats.get(), output_fmt))
    except KeyError:
        if not stats.sampler.values:
            raise SystemExit('No data received via stdin')
        raise


def main():
    argh.dispatch_command(insert_json)


if __name__ == '__main__':
    main()



import json
import aiohttp
import itertools
import calendar
import types
import time
from urllib.parse import urlparse, parse_qs
from datetime import datetime, date
from typing import List, Union, Iterable
from decimal import Decimal
from urllib.parse import urlparse, parse_qs
from cr8.aio import asyncio  # import via aio for uvloop setup

try:
    import asyncpg
except ImportError:
    asyncpg = None


HTTP_DEFAULT_HDRS = {'Content-Type': 'application/json'}


class CrateJsonEncoder(json.JSONEncoder):

    epoch = datetime(1970, 1, 1)

    def default(self, o):
        if isinstance(o, Decimal):
            return str(o)
        if isinstance(o, datetime):
            delta = o - self.epoch
            return int(delta.total_seconds() * 1000)
        if isinstance(o, date):
            return calendar.timegm(o.timetuple()) * 1000
        return json.JSONEncoder.default(self, o)


class SqlException(Exception):

    def __init__(self, message):
        self.message = message


client_errors = [
    SqlException,
    aiohttp.ClientError
]
if asyncpg:
    client_errors.append(asyncpg.exceptions.PostgresError)


def _to_http_uri(s: str) -> str:
    """Prefix the string with 'http://' if there is no schema."""
    if not s.startswith(('http://', 'https://')):
        return 'http://' + s
    return s


def _to_http_hosts(hosts: Union[Iterable[str], str]) -> List[str]:
    """Convert a string of whitespace or comma separated hosts into a list of hosts.

    Hosts may also already be a list or other iterable.
    Each host will be prefixed with 'http://' if it is not already there.

    >>> _to_http_hosts('n1:4200,n2:4200')
    ['http://n1:4200', 'http://n2:4200']

    >>> _to_http_hosts('n1:4200 n2:4200')
    ['http://n1:4200', 'http://n2:4200']

    >>> _to_http_hosts('https://n1:4200')
    ['https://n1:4200']

    >>> _to_http_hosts(['http://n1:4200', 'n2:4200'])
    ['http://n1:4200', 'http://n2:4200']
    """
    if isinstance(hosts, str):
        hosts = hosts.replace(',', ' ').split()
    return [_to_http_uri(i) for i in hosts]


async def _exec(session, url, data):
    async with session.post(url,
                            data=data,
                            headers=HTTP_DEFAULT_HDRS,
                            timeout=None) as resp:
        if resp.status == 401:
            t = await resp.text()
            raise SqlException(t)
        r = await resp.json()
        if 'error' in r:
            raise SqlException(
                r['error']['message'] + ' occurred using: ' + str(data))
        return r


def _plain_or_callable(obj):
    """Returns the value of the called object of obj is a callable,
    otherwise the plain object.
    Returns None if obj is None.

    >>> obj = None
    >>> _plain_or_callable(obj)

    >>> stmt = 'select * from sys.nodes'
    >>> _plain_or_callable(stmt)
    'select * from sys.nodes'

    >>> def _args():
    ...     return [1, 'name']
    >>> _plain_or_callable(_args)
    [1, 'name']

    >>> _plain_or_callable((x for x in range(10)))
    0

    >>> class BulkArgsGenerator:
    ...     def __call__(self):
    ...         return [[1, 'foo'], [2, 'bar'], [3, 'foobar']]
    >>> _plain_or_callable(BulkArgsGenerator())
    [[1, 'foo'], [2, 'bar'], [3, 'foobar']]
    """
    if callable(obj):
        return obj()
    elif isinstance(obj, types.GeneratorType):
        return next(obj)
    else:
        return obj


def _date_or_none(d: str) -> str:
    """Return a date as if, if valid, otherwise None

    >>> _date_or_none('2017-02-27')
    '2017-02-27'

    >>> _date_or_none('NA')
    """
    try:
        datetime.strptime(d, '%Y-%m-%d')
        return d
    except ValueError:
        return None


def _to_dsn(hosts):
    """Convert a host URI into a dsn for aiopg.

    >>> _to_dsn('aiopg://myhostname:4242/mydb')
    'postgres://crate@myhostname:4242/mydb'

    >>> _to_dsn('aiopg://myhostname:4242')
    'postgres://crate@myhostname:4242/doc'

    >>> _to_dsn('aiopg://hoschi:pw@myhostname:4242/doc?sslmode=require')
    'postgres://hoschi:pw@myhostname:4242/doc?sslmode=require'

    >>> _to_dsn('aiopg://myhostname')
    'postgres://crate@myhostname:5432/doc'
    """
    p = urlparse(hosts)
    try:
        user_and_pw, netloc = p.netloc.split('@', maxsplit=1)
    except ValueError:
        netloc = p.netloc
        user_and_pw = 'crate'
    try:
        host, port = netloc.split(':', maxsplit=1)
    except ValueError:
        host = netloc
        port = 5432
    dbname = p.path[1:] if p.path else 'doc'
    dsn = f'postgres://{user_and_pw}@{host}:{port}/{dbname}'
    if p.query:
        dsn += '?' + '&'.join(k + '=' + v[0] for k, v in parse_qs(p.query).items())
    return dsn


def _to_boolean(v):
    if str(v).lower() in ("true"):
        return True
    elif str(v).lower() in ("false"):
        return False
    else:
        raise ValueError('not a boolean value')


def _verify_ssl_from_first(hosts):
    """Check if SSL validation parameter is passed in URI

    >>> _verify_ssl_from_first(['https://myhost:4200/?verify_ssl=false'])
    False

    >>> _verify_ssl_from_first(['https://myhost:4200/'])
    True

    >>> _verify_ssl_from_first([
    ...      'https://h1:4200/?verify_ssl=False',
    ...      'https://h2:4200/?verify_ssl=True'
    ... ])
    False
    """
    for host in hosts:
        query = parse_qs(urlparse(host).query)
        if 'verify_ssl' in query:
            return _to_boolean(query['verify_ssl'][0])
    return True


class AsyncpgClient:
    def __init__(self, hosts, pool_size=25):
        self.dsn = _to_dsn(hosts)
        self.pool_size = pool_size
        self._pool = None

    async def _get_pool(self):
        if not self._pool:
            self._pool = await asyncpg.create_pool(
                self.dsn,
                min_size=self.pool_size,
                max_size=self.pool_size
            )
        return self._pool

    async def execute(self, stmt, args=None):
        start = time.time()
        pool = await self._get_pool()
        async with pool.acquire() as conn:
            if args:
                rows = await conn.fetch(stmt, *args)
            else:
                rows = await conn.fetch(stmt)
            return {
                'duration': (time.time() - start) * 1000.,
                'rows': rows
            }

    async def execute_many(self, stmt, bulk_args):
        start = time.time()
        pool = await self._get_pool()
        async with pool.acquire() as conn:
            await conn.executemany(stmt, bulk_args)
            return {
                'duration': (time.time() - start) * 1000.,
                'rows': []
            }

    async def get_server_version(self):
        pool = await self._get_pool()
        async with pool.acquire() as conn:
            for (version,) in await conn.fetch('select version from sys.nodes'):
                version = json.loads(version)
                return {
                    'hash': version['build_hash'],
                    'number': version['number']
                }

    async def _close_pool(self):
        if self._pool:
            await self._pool.close()
            self._pool = None

    def close(self):
        asyncio.get_event_loop().run_until_complete(self._close_pool())

    def __enter__(self):
        return self

    def __exit__(self, *exs):
        self.close()


class HttpClient:
    def __init__(self, hosts, conn_pool_limit=25):
        self.hosts = hosts
        self.urls = itertools.cycle([i + '/_sql' for i in hosts])
        verify_ssl = _verify_ssl_from_first(self.hosts)
        conn = aiohttp.TCPConnector(limit=conn_pool_limit, verify_ssl=verify_ssl)
        self._session = aiohttp.ClientSession(connector=conn)

    async def execute(self, stmt, args=None):
        payload = {'stmt': _plain_or_callable(stmt)}
        if args:
            payload['args'] = _plain_or_callable(args)
        return await _exec(
            self._session,
            next(self.urls),
            json.dumps(payload, cls=CrateJsonEncoder)
        )

    async def execute_many(self, stmt, bulk_args):
        data = json.dumps(dict(
            stmt=_plain_or_callable(stmt),
            bulk_args=_plain_or_callable(bulk_args)
        ), cls=CrateJsonEncoder)
        return await _exec(self._session, next(self.urls), data)

    async def get_server_version(self):
        async with self._session.get(self.hosts[0] + '/') as resp:
            r = await resp.json()
            version = r['version']
            return {
                'hash': version['build_hash'],
                'number': version['number'],
                'date': _date_or_none(version['build_timestamp'][:10])
            }

    async def _close(self):
        if self._session:
            await self._session.close()

    def close(self):
        asyncio.get_event_loop().run_until_complete(self._close())

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


def client(hosts, concurrency=25):
    hosts = hosts or 'localhost:4200'
    if hosts.startswith('asyncpg://'):
        if not asyncpg:
            raise ValueError('Cannot use "asyncpg" scheme if asyncpg is not available')
        return AsyncpgClient(hosts, pool_size=concurrency)
    return HttpClient(_to_http_hosts(hosts), conn_pool_limit=concurrency)



"""misc functions that have no real home."""

import logging
import gzip
from typing import Tuple, Iterator, Any
from collections import defaultdict


def init_logging(log):
    log.setLevel(logging.INFO)
    ch = logging.StreamHandler()
    ch.setLevel(logging.INFO)
    log.addHandler(ch)


def try_len(o: Any) -> int:
    """ Return len of `o` or None if `o` doesn't support len

    >>> try_len([1, 2])
    2

    >>> try_len(print)
    >>> try_len(None)
    """
    if not o:
        return None
    try:
        return len(o)
    except TypeError:
        return None


def parse_version(version: str) -> tuple:
    """Parse a string formatted X.Y.Z version number into a tuple

    >>> parse_version('10.2.3')
    (10, 2, 3)
    """
    if not version:
        return None
    major, minor, patch = version.split('.', maxsplit=3)
    return (int(major), int(minor), int(patch))


def parse_table(fq_table: str) -> Tuple[str, str]:
    """Parse a tablename into tuple(<schema>, <table>).

    Schema defaults to doc if the table name doesn't contain a schema.

    >>> parse_table('x.users')
    ('x', 'users')

    >>> parse_table('users')
    ('doc', 'users')
    """
    parts = fq_table.split('.')
    if len(parts) == 1:
        return 'doc', parts[0]
    elif len(parts) == 2:
        return parts[0], parts[1]
    else:
        raise ValueError


def as_bulk_queries(queries, bulk_size):
    """Group a iterable of (stmt, args) by stmt into (stmt, bulk_args).

    bulk_args will be a list of the args grouped by stmt.

    len(bulk_args) will be <= bulk_size
    """
    stmt_dict = defaultdict(list)
    for stmt, args in queries:
        bulk_args = stmt_dict[stmt]
        bulk_args.append(args)
        if len(bulk_args) == bulk_size:
            yield stmt, bulk_args
            del stmt_dict[stmt]
    for stmt, bulk_args in stmt_dict.items():
        yield stmt, bulk_args


def get_lines(filename: str) -> Iterator[str]:
    """Create an iterator that returns the lines of a utf-8 encoded file."""
    if filename.endswith('.gz'):
        with gzip.open(filename, 'r') as f:
            for line in f:
                yield line.decode('utf-8')
    else:
        with open(filename, 'r', encoding='utf-8') as f:
            for line in f:
                yield line


def as_statements(lines: Iterator[str]) -> Iterator[str]:
    """Create an iterator that transforms lines into sql statements.

    Statements within the lines must end with ";"
    The last statement will be included even if it does not end in ';'

    >>> list(as_statements(['select * from', '-- comments are filtered', 't;']))
    ['select * from t']

    >>> list(as_statements(['a;', 'b', 'c;', 'd', ' ']))
    ['a', 'b c', 'd']
    """
    lines = (l.strip() for l in lines if l)
    lines = (l for l in lines if l and not l.startswith('--'))
    parts = []
    for line in lines:
        parts.append(line.rstrip(';'))
        if line.endswith(';'):
            yield ' '.join(parts)
            parts.clear()
    if parts:
        yield ' '.join(parts)


def break_iterable(iterable, pred):
    """Break a iterable on the item that matches the predicate into lists.

    The item that matched the predicate is not included in the result.

    >>> list(break_iterable([1, 2, 3, 4], lambda x: x == 3))
    [[1, 2], [4]]
    """
    sublist = []
    for i in iterable:
        if pred(i):
            yield sublist
            sublist = []
        else:
            sublist.append(i)
    yield sublist



from unittest import TestCase
from cr8.log import format_stats
from cr8.metrics import Stats


class ResultTest(TestCase):

    def test_short_result_output_with_only_1_measurement(self):
        stats = Stats()
        stats.measure(23.4)
        self.assertEqual(
            format_stats(stats.get(), 'short'),
            ('Runtime (in ms):\n'
             '    mean:    23.400 ± 0.000')
        )

    def test_short_result_output_with_more_measurements(self):
        stats = Stats()
        stats.measure(23.4)
        stats.measure(48.7)
        stats.measure(32.5)
        stats.measure(15.9)
        self.assertEqual(
            format_stats(stats.get(), 'short'),
            ('Runtime (in ms):\n'
             '    mean:    30.125 ± 13.839\n'
             '    min/max: 15.900 → 48.700\n'
             'Percentile:\n'
             '    50:   23.400 ± 14.121 (stdev)\n'
             '    95:   48.700\n'
             '    99.9: 48.700')
        )



import argh
import os
import itertools
from functools import partial

from cr8 import aio, clients
from .insert_json import to_insert
from .bench_spec import load_spec
from .engine import Runner, Result, run_and_measure, eval_fail_if
from .misc import (
    as_bulk_queries,
    as_statements,
    get_lines,
    parse_version,
    try_len
)
from .cli import dicts_from_lines
from .log import Logger


BENCHMARK_TABLE = '''
create table if not exists benchmarks (
    version_info object (strict) as (
        number string,
        hash string,
        date timestamp
    ),
    statement string,
    meta object as (
        name string
    ),
    started timestamp,
    ended timestamp,
    concurrency int,
    bulk_size int,
    runtime_stats object (strict) as (
        avg double,
        min double,
        max double,
        mean double,
        error_margin double,
        median double,
        percentile object as (
            "50" double,
            "75" double,
            "90" double,
            "99" double,
            "99_9" double
        ),
        n integer,
        variance double,
        stdev double,
        samples array(double)
    )
) clustered into 8 shards with (number_of_replicas = '1-3', column_policy='strict')
'''


def _result_to_crate(log, client):
    table_created = []

    def save_result(result):
        if not table_created:
            aio.run(client.execute, BENCHMARK_TABLE)
            table_created.append(None)
        stmt, args = to_insert('benchmarks', result.as_dict())
        aio.run(client.execute, stmt, args)
        log.result(result)

    return save_result


class Executor:
    def __init__(self,
                 spec_dir,
                 benchmark_hosts,
                 result_hosts,
                 log,
                 fail_if,
                 sample_mode):
        self.benchmark_hosts = benchmark_hosts
        self.sample_mode = sample_mode
        self.spec_dir = spec_dir
        self.client = clients.client(benchmark_hosts)
        self.result_client = clients.client(result_hosts)
        self.server_version_info = aio.run(self.client.get_server_version)
        self.server_version = parse_version(self.server_version_info['number'])
        self.log = log
        self.create_result = partial(
            Result,
            version_info=self.server_version_info
        )
        if fail_if:
            self.fail_if = partial(eval_fail_if, fail_if)
        else:
            self.fail_if = lambda x: None
        if result_hosts:
            self.process_result = _result_to_crate(self.log, self.result_client)
        else:
            self.process_result = log.result

    def _to_inserts(self, data_spec):
        target = data_spec['target']
        source = os.path.join(self.spec_dir, data_spec['source'])
        dicts = dicts_from_lines(get_lines(source))
        return (to_insert(target, d) for d in dicts)

    def exec_instructions(self, instructions):
        filenames = instructions.statement_files
        filenames = (os.path.join(self.spec_dir, i) for i in filenames)
        lines = (line for fn in filenames for line in get_lines(fn))
        statements = itertools.chain(as_statements(lines), instructions.statements)
        for stmt in statements:
            aio.run(self.client.execute, stmt)

        for data_file in instructions.data_files:
            inserts = as_bulk_queries(self._to_inserts(data_file),
                                      data_file.get('bulk_size', 5000))
            concurrency = data_file.get('concurrency', 25)
            aio.run_many(self.client.execute_many, inserts, concurrency=concurrency)
            aio.run(self.client.execute, 'refresh table {target}'.format(target=data_file['target']))

    def run_load_data(self, data_spec, meta=None):
        inserts = self._to_inserts(data_spec)
        statement = next(iter(inserts))[0]
        bulk_size = data_spec.get('bulk_size', 5000)
        inserts = as_bulk_queries(self._to_inserts(data_spec), bulk_size)
        concurrency = data_spec.get('concurrency', 25)
        num_records = data_spec.get('num_records')
        if num_records:
            num_records = max(1, int(num_records / bulk_size))
        timed_stats = run_and_measure(
            self.client.execute_many, inserts, concurrency, num_records)
        self.process_result(self.create_result(
            statement=statement,
            meta=meta,
            timed_stats=timed_stats,
            concurrency=concurrency,
            bulk_size=bulk_size,
        ))

    def _skip_message(self, min_version, stmt):
        msg = ('## Skipping (Version {server_version} instead of {min_version}):\n'
               '   Statement: {statement:.70}')
        msg = msg.format(
            statement=stmt,
            min_version='.'.join((str(x) for x in min_version)),
            server_version='.'.join((str(x) for x in self.server_version)))
        return msg

    def run_queries(self, queries, meta=None):
        for query in queries:
            stmt = query['statement']
            iterations = query.get('iterations', 1)
            duration = query.get('duration', None)
            concurrency = query.get('concurrency', 1)
            args = query.get('args')
            bulk_args = query.get('bulk_args')
            min_version = parse_version(query.get('min_version'))
            if min_version and min_version > self.server_version:
                self.log.info(self._skip_message(min_version, stmt))
                continue
            mode_desc = 'Duration' if duration else 'Iterations'
            self.log.info(
                (f'\n## Running Query:\n'
                 f'   Statement:\n'
                 f'     {stmt}\n'
                 f'   Concurrency: {concurrency}\n'
                 f'   {mode_desc}: {duration or iterations}')
            )
            with Runner(self.benchmark_hosts, concurrency, self.sample_mode) as runner:
                timed_stats = runner.run(
                    stmt,
                    iterations=iterations,
                    duration=duration,
                    args=args,
                    bulk_args=bulk_args
                )
            result = self.create_result(
                statement=stmt,
                meta=meta,
                timed_stats=timed_stats,
                concurrency=concurrency,
                bulk_size=try_len(bulk_args)
            )
            self.process_result(result)
            self.fail_if(result)

    def __enter__(self):
        return self

    def __exit__(self, *ex):
        self.client.close()
        self.result_client.close()


def do_run_spec(spec,
                benchmark_hosts,
                *,
                log,
                sample_mode,
                result_hosts=None,
                action=None,
                fail_if=None):
    with Executor(
        spec_dir=os.path.dirname(spec),
        benchmark_hosts=benchmark_hosts,
        result_hosts=result_hosts,
        log=log,
        fail_if=fail_if,
        sample_mode=sample_mode
    ) as executor:
        spec = load_spec(spec)
        try:
            if not action or 'setup' in action:
                log.info('# Running setUp')
                executor.exec_instructions(spec.setup)
            log.info('# Running benchmark')
            if spec.load_data and (not action or 'load_data' in action):
                for data_spec in spec.load_data:
                    executor.run_load_data(data_spec, spec.meta)
            if spec.queries and (not action or 'queries' in action):
                executor.run_queries(spec.queries, spec.meta)
        finally:
            if not action or 'teardown' in action:
                log.info('# Running tearDown')
                executor.exec_instructions(spec.teardown)


@argh.arg('benchmark_hosts', type=str)
@argh.arg('-of', '--output-fmt', choices=['json', 'text'], default='text')
@argh.arg('--action',
          choices=['setup', 'teardown', 'queries', 'load_data'],
          action='append')
@argh.arg('--logfile-info', help='Redirect info messages to a file')
@argh.arg('--logfile-result', help='Redirect benchmark results to a file')
@argh.arg('--sample-mode', choices=('all', 'reservoir'),
          help='Method used for sampling', default='reservoir')
@argh.wrap_errors([KeyboardInterrupt, BrokenPipeError] + clients.client_errors)
def run_spec(spec,
             benchmark_hosts,
             result_hosts=None,
             output_fmt=None,
             logfile_info=None,
             logfile_result=None,
             action=None,
             fail_if=None,
             sample_mode='reservoir'):
    """Run a spec file, executing the statements on the benchmark_hosts.

    Short example of a spec file:

        [setup]
        statement_files = ["sql/create_table.sql"]

            [[setup.data_files]]
            target = "t"
            source = "data/t.json"

        [[queries]]
        statement = "select count(*) from t"
        iterations = 2000
        concurrency = 10

        [teardown]
        statements = ["drop table t"]

    See https://github.com/mfussenegger/cr8/tree/master/specs
    for more examples.

    Args:
        spec: path to a spec file
        benchmark_hosts: hostname[:port] pairs of Crate nodes
        result_hosts: optional hostname[:port] Crate node pairs into which the
            runtime statistics should be inserted.
        output_fmt: output format
        action: Optional action to execute.
            Default is to execute all actions - setup, queries and teardown.
            If present only the specified action will be executed.
            The argument can be provided multiple times to execute more than
            one action.
        fail-if: An expression that causes cr8 to exit with a failure if it
            evaluates to true.
            The expression can contain formatting expressions for:
                - runtime_stats
                - statement
                - meta
                - concurrency
                - bulk_size
            For example:
                --fail-if "{runtime_stats.mean} > 1.34"
    """
    with Logger(output_fmt=output_fmt,
                logfile_info=logfile_info,
                logfile_result=logfile_result) as log:
        do_run_spec(
            spec=spec,
            benchmark_hosts=benchmark_hosts,
            log=log,
            result_hosts=result_hosts,
            action=action,
            fail_if=fail_if,
            sample_mode=sample_mode
        )


def main():
    argh.dispatch_command(run_spec)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import json
import argh
import argparse
import operator
import math
import signal
from faker import Factory
from functools import partial
from collections import OrderedDict
from concurrent.futures import ThreadPoolExecutor

from .insert_json import to_insert
from .misc import parse_table, parse_version
from .aio import asyncio, consume
from .cli import to_int
from .fake_providers import GeoSpatialProvider, auto_inc
from cr8 import clients, aio

loop = asyncio.get_event_loop()


SELLECT_COLS = """
select
    column_name,
    data_type
from
    information_schema.columns
where
    is_generated = false
    and {schema_column_name} = ?
    and table_name = ?
    and column_name not like '%]'
order by ordinal_position asc
"""


def retrieve_columns(client, schema, table):
    r = aio.run(client.execute, "select min(version['number']) from sys.nodes")
    version = parse_version(r['rows'][0][0])
    stmt = SELLECT_COLS.format(
        schema_column_name='table_schema' if version >= (0, 57, 0) else 'schema_name')
    r = aio.run(client.execute, stmt, (schema, table))
    return OrderedDict(r['rows'])


def generate_row(fakers):
    return [x() for x in fakers]


def array_provider(len_provider, value_provider, dimensions):
    if dimensions == 0:
        return value_provider()
    else:
        return [array_provider(len_provider, value_provider, dimensions - 1)
                for _ in range(len_provider())]


def make_array_provider(inner_provider, dimensions):
    def setup_array_providers(fake):
        inner = inner_provider(fake)
        arr_len = partial(fake.random_int, min=0, max=50)
        return partial(array_provider, arr_len, inner, dimensions)
    return setup_array_providers


class DataFaker:
    _mapping = {
        ('id', 'string'): operator.attrgetter('uuid4'),
        ('id', 'integer'): auto_inc,
        ('id', 'long'): auto_inc,
    }

    _type_default = {
        'byte': lambda f: partial(f.random_int, min=-128, max=127),
        'short': lambda f: partial(f.random_int, min=-32768, max=32767),
        'integer': lambda f: partial(
            f.random_int, min=-2147483648, max=2147483647),
        'long': lambda f: partial(
            f.random_int, min=-9223372036854775808, max=9223372036854775807),
        'float': operator.attrgetter('pyfloat'),
        'double': operator.attrgetter('pydecimal'),
        'ip': operator.attrgetter('ipv4'),
        'timestamp': lambda f: partial(
            f.date_time_between, start_date='-2y', end_date='now'),
        'string': operator.attrgetter('word'),
        'boolean': operator.attrgetter('boolean'),
        'geo_point': operator.attrgetter('geo_point'),
        'geo_shape': operator.attrgetter('geo_shape'),
        'object': lambda f: dict
    }

    _custom = {
        'auto_inc': auto_inc
    }

    def __init__(self):
        self.fake = Factory.create()
        self.fake.add_provider(GeoSpatialProvider)

    def _provider_for_type(self, data_type):
        inner_type, *dim = data_type.split('_array')
        inner_provider = self._type_default.get(inner_type)
        if not dim or not inner_provider:
            return inner_provider
        return make_array_provider(inner_provider, len(dim))

    def provider_for_column(self, column_name, data_type):
        provider = getattr(self.fake, column_name, None)
        if provider:
            return provider
        custom_provider = self._custom.get(column_name)
        if custom_provider:
            return custom_provider(self.fake)
        alternative = self._mapping.get((column_name, data_type))
        if not alternative:
            alternative = self._provider_for_type(data_type)
            if not alternative:
                msg = 'No fake provider found for column "{col}" with type "{type}"'
                raise ValueError(msg.format(col=column_name, type=data_type))
        return alternative(self.fake)

    def provider_from_mapping(self, column_name, mapping):
        key = mapping[column_name]
        args = None
        if isinstance(key, list):
            key, args = key
        provider = getattr(self.fake, key, None)
        if not provider:
            raise KeyError('No fake provider with name "%s" found' % (key,))
        if args:
            provider = partial(provider, *args)
        return provider


def create_row_generator(columns, mapping=None):
    fake = DataFaker()
    fakers = []
    for column_name, type_name in columns.items():
        if mapping and column_name in mapping:
            fakers.append(fake.provider_from_mapping(column_name, mapping))
        else:
            fakers.append(fake.provider_for_column(column_name, type_name))
    return partial(generate_row, fakers)


async def _exec_many(client, stmt, args_coro):
    return await client.execute_many(stmt, await args_coro)


def _create_bulk_args(row_fun, req_size):
    return [row_fun() for i in range(req_size)]


def _bulk_size_generator(num_records, bulk_size, active):
    """ Generate bulk_size until num_records is reached or active becomes false

    >>> gen = _bulk_size_generator(155, 50, [True])
    >>> list(gen)
    [50, 50, 50, 5]
    """
    while active and num_records > 0:
        req_size = min(num_records, bulk_size)
        num_records -= req_size
        yield req_size


async def _gen_data_and_insert(q, e, client, stmt, row_fun, size_seq):
    for size in size_seq:
        args_coro = loop.run_in_executor(e, _create_bulk_args, row_fun, size)
        task = asyncio.ensure_future(_exec_many(client, stmt, args_coro))
        await q.put(task)
    await q.put(None)


@argh.arg('--table', help='table name', required=True)
@argh.arg('--hosts', help='crate hosts', type=str)
@argh.arg('-n', '--num-records',
          help='number of records to insert',
          type=to_int,
          default=int(1e5))
@argh.arg('-b', '--bulk-size', type=to_int)
@argh.arg('-c', '--concurrency', type=to_int)
@argh.arg('--mapping-file',
          type=argparse.FileType('r'),
          help='JSON file with a column to fake provider mapping.')
@argh.wrap_errors([KeyboardInterrupt] + clients.client_errors)
def insert_fake_data(hosts=None,
                     table=None,
                     num_records=1e5,
                     bulk_size=1000,
                     concurrency=25,
                     mapping_file=None):
    """Generate random data and insert it into a table.

    This will read the table schema and then find suitable random data providers.
    Which provider is choosen depends on the column name and data type.

    Example:

        A column named `name` will map to the `name` provider.
        A column named `x` of type int will map to `random_int` because there
        is no `x` provider.

    Available providers are listed here:
        https://faker.readthedocs.io/en/latest/providers.html

        Additional providers:
        - auto_inc:
            Returns unique incrementing numbers.
            Automatically used for columns named "id" of type int or long
        - geo_point
            Returns [<lon>, <lat>]
            Automatically used for columns of type geo_point

    Args:
        hosts: <host>:[<port>] of the Crate node
        table: The table name into which the data should be inserted.
            Either fully qualified: `<schema>.<table>` or just `<table>`
        num_records: Number of records to insert.
            Usually a number but expressions like `1e4` work as well.
        bulk_size: The bulk size of the insert statements.
        concurrency: How many operations to run concurrently.
        mapping_file: A JSON file that defines a mapping from column name to
            fake-factory provider.
            The format is as follows:
            {
                "column_name": ["provider_with_args", ["arg1", "arg"]],
                "x": ["provider_with_args", ["arg1"]],
                "y": "provider_without_args"
            }
    """
    with clients.client(hosts, concurrency=1) as client:
        schema, table_name = parse_table(table)
        columns = retrieve_columns(client, schema, table_name)
    if not columns:
        sys.exit('Could not find columns for table "{}"'.format(table))
    print('Found schema: ')
    print(json.dumps(columns, sort_keys=True, indent=4))
    mapping = None
    if mapping_file:
        mapping = json.load(mapping_file)

    bulk_size = min(num_records, bulk_size)
    num_inserts = int(math.ceil(num_records / bulk_size))

    gen_row = create_row_generator(columns, mapping)

    stmt = to_insert('"{schema}"."{table_name}"'.format(**locals()), columns)[0]
    print('Using insert statement: ')
    print(stmt)

    print('Will make {} requests with a bulk size of {}'.format(
        num_inserts, bulk_size))

    print('Generating fake data and executing inserts')
    q = asyncio.Queue(maxsize=concurrency)
    with clients.client(hosts, concurrency=concurrency) as client:
        active = [True]

        def stop():
            asyncio.ensure_future(q.put(None))
            active.clear()
            loop.remove_signal_handler(signal.SIGINT)
        if sys.platform != 'win32':
            loop.add_signal_handler(signal.SIGINT, stop)
        bulk_seq = _bulk_size_generator(num_records, bulk_size, active)
        with ThreadPoolExecutor() as e:
            tasks = asyncio.gather(
                _gen_data_and_insert(q, e, client, stmt, gen_row, bulk_seq),
                consume(q, total=num_inserts)
            )
            loop.run_until_complete(tasks)


def main():
    argh.dispatch_command(insert_fake_data)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import argh

from cr8 import aio
from cr8.cli import lines_from_stdin, to_int
from cr8.misc import as_statements
from cr8.log import Logger
from cr8.clients import client_errors
from cr8.engine import Runner, Result, eval_fail_if


@argh.arg('--hosts', help='crate hosts', type=str)
@argh.arg('-s', '--stmt', type=str)
@argh.arg('-w', '--warmup', type=to_int)
@argh.arg('-r', '--repeat', type=to_int)
@argh.arg('--duration',
          type=to_int,
          help='Duration in seconds. Overwrites repeat if set.')
@argh.arg('-c', '--concurrency', type=to_int)
@argh.arg('-of', '--output-fmt', choices=['json', 'text'], default='text')
@argh.arg('--fail-if', help='An expression which causes cr8 to exit with a\
          failure if it evaluates to true')
@argh.arg('--sample-mode', choices=('all', 'reservoir'),
          help='Method used for sampling', default='reservoir')
@argh.wrap_errors([KeyboardInterrupt, BrokenPipeError] + client_errors)
def timeit(hosts=None,
           stmt=None,
           warmup=30,
           repeat=None,
           duration=None,
           concurrency=1,
           output_fmt=None,
           fail_if=None,
           sample_mode='reservoir'):
    """Run the given statement a number of times and return the runtime stats

    Args:
        fail-if: An expression that causes cr8 to exit with a failure if it
            evaluates to true.
            The expression can contain formatting expressions for:
                - runtime_stats
                - statement
                - meta
                - concurrency
                - bulk_size
            For example:
                --fail-if "{runtime_stats.mean} > 1.34"
    """
    num_lines = 0
    log = Logger(output_fmt)
    with Runner(hosts, concurrency, sample_mode) as runner:
        version_info = aio.run(runner.client.get_server_version)
        for line in as_statements(lines_from_stdin(stmt)):
            runner.warmup(line, warmup)
            timed_stats = runner.run(line, iterations=repeat, duration=duration)
            r = Result(
                version_info=version_info,
                statement=line,
                timed_stats=timed_stats,
                concurrency=concurrency
            )
            log.result(r)
            if fail_if:
                eval_fail_if(fail_if, r)
        num_lines += 1
    if num_lines == 0:
        raise SystemExit(
            'No SQL statements provided. Use --stmt or provide statements via stdin')


def main():
    argh.dispatch_command(timeit)


if __name__ == '__main__':
    main()




import os
from os.path import basename
import json
import toml


class Instructions:
    def __init__(self,
                 statements=None,
                 statement_files=None,
                 data_files=None):
        self.statements = statements or []
        self.statement_files = statement_files or []
        self.data_files = data_files or []

    @staticmethod
    def from_dict(d):
        return Instructions(
            statements=d.get('statements', []),
            statement_files=d.get('statement_files', []),
            data_files=d.get('data_files', [])
        )


class Spec:
    def __init__(self, setup, teardown, queries=None, load_data=None, meta=None):
        self.setup = setup
        self.teardown = teardown
        self.queries = queries
        self.load_data = load_data
        self.meta = meta or {}

    @staticmethod
    def from_dict(d):
        return Spec(
            setup=Instructions.from_dict(d.get('setup', {})),
            teardown=Instructions.from_dict(d.get('teardown', {})),
            meta=d.get('meta', {}),
            queries=d.get('queries', []),
            load_data=d.get('load_data', []),
        )

    @staticmethod
    def from_json_file(filename):
        with open(filename, 'r', encoding='utf-8') as spec_file:
            spec = Spec.from_dict(json.load(spec_file))
            spec.meta.setdefault('name', basename(filename))
            return spec

    @staticmethod
    def from_toml_file(filename):
        with open(filename, 'r', encoding='utf-8') as spec_file:
            spec = Spec.from_dict(toml.loads(spec_file.read()))
            spec.meta.setdefault('name', basename(filename))
            return spec

    @staticmethod
    def from_python_file(filename):
        with open(filename, 'r', encoding='utf-8') as f:
            global_vars = {
                'Spec': Spec,
                'Instructions': Instructions
            }
            code = compile(f.read(), filename, 'exec')
            exec(code, global_vars)
            spec = global_vars['spec']
            spec.meta.setdefault('name', basename(filename))
            return spec


spec_loaders = {
    '.json': Spec.from_json_file,
    '.toml': Spec.from_toml_file,
    '.py': Spec.from_python_file
}


def load_spec(spec_file):
    ext = os.path.splitext(spec_file)[1]
    loader = spec_loaders[ext]
    return loader(spec_file)




from unittest import TestCase, main
from doctest import DocTestSuite
from cr8 import misc


class MiscTest(TestCase):

    def test_as_bulk_queries(self):
        queries = [
            ('x', (1, 2)),
            ('x', (3, 4)),
            ('x', (5, 6)),
            ('y', (1, 2))]
        bulk_queries = sorted(list(misc.as_bulk_queries(queries, 2)))
        self.assertEqual(bulk_queries, [
            ('x', [(1, 2), (3, 4)]),
            ('x', [(5, 6)]),
            ('y', [(1, 2)])
        ])


def load_tests(loader, tests, ignore):
    tests.addTests(DocTestSuite(misc))
    return tests


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import math
from faker.providers import BaseProvider
from multiprocessing import Manager

EARTH_RADIUS = 6371  # earth radius in km

# Not implemented as Provider because with providers it's not possible to have
# 1 instance per column. So either there would be one shared Counter accross
# multiple auto_inc columns or there could be duplicate values within one column


class Counter:
    def __init__(self, value, lock):
        self.value = value
        self.lock = lock

    def __call__(self):
        val = self.value
        with self.lock:
            val.value += 1
            return val.value


def auto_inc(fake):
    manager = Manager()
    return Counter(manager.Value('i', 0), manager.Lock())


def _dest_point(point, distance, bearing, radius):
    # calculation taken from
    # https://cdn.rawgit.com/chrisveness/geodesy/v1.1.2/latlon-spherical.js
    # https://www.movable-type.co.uk/scripts/latlong.html

    δ = distance / radius  # angular distance in rad
    θ = math.radians(bearing)

    φ1 = math.radians(point[1])
    λ1 = math.radians(point[0])

    sinφ1 = math.sin(φ1)
    cosφ1 = math.cos(φ1)
    sinδ = math.sin(δ)
    cosδ = math.cos(δ)
    sinθ = math.sin(θ)
    cosθ = math.cos(θ)

    sinφ2 = sinφ1 * cosδ + cosφ1 * sinδ * cosθ
    φ2 = math.asin(sinφ2)
    y = sinθ * sinδ * cosφ1
    x = cosδ - sinφ1 * sinφ2
    λ2 = λ1 + math.atan2(y, x)

    return [
        (math.degrees(λ2) + 540) % 360 - 180,  # normalise to −180..+180°
        math.degrees(φ2)
    ]


class GeoSpatialProvider(BaseProvider):
    """
    A Faker provider for geospatial data types, such as GEO_POINT.

    >>> from faker import Faker
    >>> fake = Faker()
    >>> fake.add_provider(GeoSpatialProvider)
    >>> hasattr(fake, 'geo_point')
    True

    >>> loc = fake.geo_point()
    >>> -180 <= loc[0] <= 180
    True
    >>> -90 <= loc[1] <= 90
    True

    >>> loc = fake.geo_point(0.0, 180.0, 0.0, 90.0)
    >>> 0 <= loc[0] <= 180
    True
    >>> 0 <= loc[1] <= 90
    True
    """

    def geo_point(self,
                  lon_min=-180.0, lon_max=180.0,
                  lat_min=-90.0, lat_max=90.0):
        assert isinstance(lon_min, float)
        assert isinstance(lon_max, float)
        assert isinstance(lat_min, float)
        assert isinstance(lat_max, float)

        assert lon_min >= -180.0
        assert lon_max <= 180.0
        assert lat_min >= -90.0
        assert lat_max <= 90.0

        # longitude: -180 .. 0 .. +180 (E-W)
        # latitude: -90 .. 0 .. +90 (S-N)
        u = self.generator.random.uniform
        return [
            u(lon_min, lon_max),
            u(lat_min, lat_max)
        ]

    def geo_shape(self, sides=5, center=None, distance=None):
        """
        Return a WKT string for a POLYGON with given amount of sides.
        The polygon is defined by its center (random point if not provided) and
        the distance (random distance if not provided; in km) of the points to
        its center.
        """
        assert isinstance(sides, int)

        if distance is None:
            distance = self.random_int(100, 1000)
        else:
            # 6371 => earth radius in km
            # assert that shape radius is maximum half of earth's circumference
            assert isinstance(distance, int)
            assert distance <= EARTH_RADIUS * math.pi, \
                'distance must not be greater than half of earth\'s circumference'

        if center is None:
            # required minimal spherical distance from north/southpole
            dp = distance * 180.0 / EARTH_RADIUS / math.pi
            center = self.geo_point(lat_min=-90.0 + dp, lat_max=90.0 - dp)
        else:
            assert -180.0 <= center[0] <= 180.0, 'Longitude out of bounds'
            assert -90.0 <= center[1] <= 90.0, 'Latitude out of bounds'

        angles = list(self.random_sample(range(360), sides))
        angles.sort()
        points = [_dest_point(center, distance, bearing, EARTH_RADIUS) for bearing in angles]
        # close polygon
        points.append(points[0])

        path = ', '.join([' '.join(p) for p in ([str(lon), str(lat)] for lon, lat in points)])
        return f'POLYGON (( {path} ))'



from unittest import main
from doctest import DocTestSuite
from cr8 import insert_json


def load_tests(loader, tests, ignore):
    tests.addTests(DocTestSuite(insert_json))
    return tests


if __name__ == "__main__":
    main()




from itertools import count
from cr8.bench_spec import Spec, Instructions


def queries():
    for i in range(5):
        yield {
            'statement': 'insert into t (x) values (?)',
            'args': [i]
        }
    c = count(100)
    yield {
        'statement': 'insert into t (x) values (?)',
        'bulk_args': lambda: [[next(c)] for i in range(10)]
    }


# Spec and Instructions are injected by the spec runner
spec = Spec(
    setup=Instructions(statements=["create table t (x int)"]),
    teardown=Instructions(statements=["drop table t"]),
    queries=queries(),
)



import statistics
import random
import math
from functools import partial


DEFAULT_NUM_SAMPLES = 1000


def percentile(sorted_values, p):
    """Calculate the percentile using the nearest rank method.

    >>> percentile([15, 20, 35, 40, 50], 50)
    35

    >>> percentile([15, 20, 35, 40, 50], 40)
    20

    >>> percentile([], 90)
    Traceback (most recent call last):
        ...
    ValueError: Too few data points (0) for 90th percentile
    """
    size = len(sorted_values)
    idx = (p / 100.0) * size - 0.5
    if idx < 0 or idx > size:
        raise ValueError('Too few data points ({}) for {}th percentile'.format(size, p))
    return sorted_values[int(idx)]


class UniformReservoir:
    """Reservoir Sampling Algorithm R by Jeffrey Vitter.

    See https://en.wikipedia.org/wiki/Reservoir_sampling#Algorithm_R
    """

    def __init__(self, size=DEFAULT_NUM_SAMPLES):
        self.size = size
        self.count = 0
        self.values = []

    def add(self, value):
        count = self.count
        if count < self.size:
            self.values.append(value)
        else:
            k = random.randint(0, self.count)
            if k < self.size:
                self.values[k] = value
        self.count = count + 1


class All:
    """Sampler that keeps all values"""

    def __init__(self):
        self.values = []

    def add(self, value):
        self.values.append(value)

    @property
    def count(self):
        return len(self.values)


def get_sampler(sample_mode: str):
    """Return a sampler constructor

    >>> get_sampler('all')
    <class 'cr8.metrics.All'>

    >>> get_sampler('reservoir')
    <class 'cr8.metrics.UniformReservoir'>

    >>> get_sampler('reservoir:100')
    functools.partial(<class 'cr8.metrics.UniformReservoir'>, size=100)
    """
    if sample_mode == 'all':
        return All
    mode = sample_mode.split(':')
    if mode[0] == 'reservoir':
        if len(mode) == 2:
            return partial(UniformReservoir, size=int(mode[1]))
        else:
            return UniformReservoir
    raise TypeError(f'Invalid sample_mode: {sample_mode}')


_z_map = {
    80: 1.282,
    85: 1.440,
    90: 1.645,
    95: 1.960,
    99: 2.576,
    99.5: 2.807,
    99.9: 3.291,
}


def error_margin(confidence_level, stdev, sample_size):
    return _z_map[confidence_level] * (stdev / math.sqrt(sample_size))


class Stats:
    plevels = [50, 75, 90, 95, 99, 99.9]

    def __init__(self, sampler=None):
        self.sampler = sampler() if sampler else UniformReservoir()

    def measure(self, value):
        self.sampler.add(value)

    def get(self):
        values = sorted(self.sampler.values)
        count = len(values)
        # instead of failing return empty / subset so that json2insert & co
        # don't fail
        if count == 0:
            return dict(n=0)
        elif count == 1:
            return dict(
                min=values[0],
                max=values[0],
                mean=values[0],
                n=count,
                stdev=0,
                samples=self.sampler.values
            )
        percentiles = [percentile(values, p) for p in self.plevels]
        min_ = values[0]
        max_ = values[-1]
        stdev = statistics.stdev(values)
        return dict(
            min=min_,
            max=max_,
            mean=statistics.mean(values),
            median=statistics.median(values),
            variance=statistics.variance(values),
            error_margin=error_margin(95, stdev, self.sampler.count),
            stdev=stdev,
            # replace . with _ so that the output can be inserted into crate
            # crate doesn't allow dots in column names
            percentile={str(i[0]).replace('.', '_'): i[1] for i in
                        zip(self.plevels, percentiles)},
            n=self.sampler.count,
            samples=self.sampler.values
        )



import json
from datetime import datetime, date
from decimal import Decimal
from unittest import main, TestCase
from doctest import DocTestSuite
from cr8 import clients
from cr8.clients import CrateJsonEncoder


class EncoderTest(TestCase):

    def test_decimal_encoding(self):
        s = json.dumps({'x': Decimal(10)}, cls=CrateJsonEncoder)
        self.assertEqual(s, '{"x": "10"}')

    def test_datetime_encoding(self):
        d = {'x': datetime(2017, 1, 26, 23, 33, 1)}
        s = json.dumps(d, cls=CrateJsonEncoder)
        self.assertEqual(s, '{"x": 1485473581000}')

    def test_date_encoding(self):
        d = {'x': date(2017, 1, 26)}
        s = json.dumps(d, cls=CrateJsonEncoder)
        self.assertEqual(s, '{"x": 1485388800000}')


def load_tests(loader, tests, ignore):
    tests.addTests(DocTestSuite(clients))
    return tests


if __name__ == "__main__":
    main()



import os
import doctest
import subprocess
import functools
from cr8.run_crate import CrateNode, get_crate
from crate.client import connect


crate_dir = get_crate('latest-testing')
node = CrateNode(
    crate_dir=crate_dir,
    settings={
        'cluster.name': 'cr8-tests',
        'http.port': '44200-44250',
        'bootstrap.system_call_filter': False
    })


def setup(*args):
    with connect(node.http_url) as conn:
        c = conn.cursor()
        c.execute('create table x.demo (id int, name string, country string) \
                  with (number_of_replicas = 0)')
        c.execute('create blob table blobtable with (number_of_replicas = 0)')


def teardown(*args):
    try:
        with connect(node.http_url) as conn:
            c = conn.cursor()
            c.execute('drop table x.demo')
            c.execute('drop blob table blobtable')
    finally:
        node.stop()


def transform(s):
    s = s.replace('localhost:4200', node.http_url)
    s = s.replace(
        'asyncpg://localhost:5432',
        f'asyncpg://{node.addresses.psql.host}:{node.addresses.psql.port}')
    return (
        r'print(sh("""%s""").stdout.decode("utf-8"))' % s) + '\n'


class Parser(doctest.DocTestParser):

    def parse(self, string, name='<string>'):
        r = super().parse(string, name)
        for s in r:
            if isinstance(s, doctest.Example):
                s.source = transform(s.source)
        return r


def load_tests(loader, tests, ignore):
    env = os.environ.copy()
    env['CR8_NO_TQDM'] = 'True'
    node.start()
    assert node.http_host, "http_url must be available"
    tests.addTests(doctest.DocFileSuite(
        os.path.join('..', 'README.rst'),
        globs={
            'sh': functools.partial(
                subprocess.run,
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                timeout=60,
                shell=True,
                env=env
            )
        },
        optionflags=doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS,
        setUp=setup,
        tearDown=teardown,
        parser=Parser()
    ))
    return tests


if __name__ == "__main__":
    import unittest
    unittest.main()




import functools
import os
import asyncio
import signal
import sys
try:
    import uvloop
    asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
except ImportError:
    pass

from tqdm import tqdm
tqdm = functools.partial(
    tqdm,
    unit=' requests',
    smoothing=0.1,
    disable=os.environ.get('CR8_NO_TQDM') == 'True'
)


async def measure(stats, f, *args, **kws):
    r = await f(*args, **kws)
    duration = r['duration']
    stats.measure(duration)
    return r


async def qmap(q, corof, iterable):
    for i in iterable:
        task = asyncio.ensure_future(corof(*i))
        await q.put(task)
    await q.put(None)


async def consume(q, total=None):
    last_error = None
    with tqdm(total=total) as t:
        while True:
            task = await q.get()
            if task is None:
                break
            try:
                await task
            except Exception as e:
                last_error = e
            t.update(1)
        if last_error:
            raise last_error


async def map(coro, iterable, total=None):
    for i in tqdm(iterable, total=total):
        await coro(*i)


def run(coro, *args):
    loop = asyncio.get_event_loop()
    if args:
        gen = coro(*args)
    else:
        gen = coro()
    return loop.run_until_complete(gen)


def interruptable(iterable, is_active):
    for i in iterable:
        if not is_active:
            return
        yield i


def setup_sigint_handling(loop, q, iterable):
    if sys.platform == 'win32':
        return iterable
    is_active = [True]
    iterable = interruptable(iterable, is_active)

    def stop():
        asyncio.ensure_future(q.put(None))
        is_active.clear()
        loop.remove_signal_handler(signal.SIGINT)
    loop.add_signal_handler(signal.SIGINT, stop)
    return iterable


def remove_sigint_handler(loop):
    if sys.platform == 'win32':
        return
    loop.remove_signal_handler(signal.SIGINT)


def run_many(coro, iterable, concurrency, num_items=None):
    loop = asyncio.get_event_loop()
    if concurrency == 1:
        return loop.run_until_complete(map(coro, iterable, total=num_items))
    q = asyncio.Queue(maxsize=concurrency)
    iterable = setup_sigint_handling(loop, q, iterable)
    tasks = asyncio.gather(
        qmap(q, coro, iterable),
        consume(q, total=num_items)
    )
    try:
        loop.run_until_complete(tasks)
    except KeyboardInterrupt:
        tasks.cancel()
    remove_sigint_handler(loop)



from unittest import main, TestCase
from doctest import DocTestSuite
from cr8 import run_crate
from cr8.run_crate import AddrConsumer


lines1 = [
    "[2016-06-11 19:10:16,141][INFO ][node                     ] [Ankhi] initialized",
    "[2016-06-11 19:10:16,141][INFO ][node                     ] [Ankhi] starting ...",
    "[2016-06-11 19:10:16,171][INFO ][http                     ] [Ankhi] publish_address {10.68.2.10:4200}, bound_addresses {[::]:4200}",
    "[2016-06-11 19:10:16,188][INFO ][discovery                ] [Ankhi] crate/GJAvonoFSfS3Y1IaUPTqfA"
]


lines2 = [
    "[2016-06-11 21:26:53,798][INFO ][node                     ] [Rex Mundi] starting ...",
    "[2016-06-11 21:26:53,828][INFO ][http                     ] [Rex Mundi] bound_address {inet[/0:0:0:0:0:0:0:0:4200]}, publish_address {inet[/192.168.0.19:4200]}",
]


lines3 = [
    "[2016-06-15 22:18:36,639][INFO ][node                     ] [crate] starting ...",
    "[2016-06-15 22:18:36,755][INFO ][http                     ] [crate] publish_address {localhost/127.0.0.1:42203}, bound_addresses {127.0.0.1:42203}, {[::1]:42203}, {[fe80::1]:42203}",
    "[2016-06-15 22:18:36,774][INFO ][transport                ] [crate] publish_address {localhost/127.0.0.1:4300}, bound_addresses {127.0.0.1:4300}, {[::1]:4300}, {[fe80::1]:4300}",
    "[2016-06-15 22:18:36,779][INFO ][discovery                ] [crate] Testing42203/Eroq_ZAgT4CDpF_gzh4tcA",
]

lines4 = [
    "[2016-06-16 10:27:20,074][INFO ][node                     ] [Selene] starting ...",
    "[2016-06-16 10:27:20,150][INFO ][http                     ] [Selene] bound_address {inet[/192.168.43.105:4200]}, publish_address {inet[Haudis-MacBook-Pro.local/192.168.43.105:4200]}",
    "[2016-06-16 10:27:20,165][INFO ][transport                ] [Selene] bound_address {inet[/192.168.43.105:4300]}, publish_address {inet[Haudis-MacBook-Pro.local/192.168.43.105:4300]}",
    "[2016-06-16 10:27:20,185][INFO ][discovery                ] [Selene] crate/h9moKMrATmCElYXjfad5Vw",
]

ipv6_psql = "[2017-10-26T15:31:44,654][INFO ][psql                     ] [Deichselspitze] publish_address {[::1]:5432}, bound_addresses {[::1]:5432}"
ipv6_http = "[2017-10-26T15:31:44,666][INFO ][i.c.p.h.CrateNettyHttpServerTransport] [Deichselspitze] publish_address {[::1]:44200}, bound_addresses {[::1]:44200}"

crate12_psql = "[2017-02-01T15:43:37,736][INFO ][psql                     ] [8f64DTi] publish_address {127.0.0.1:5432}, bound_addresses {127.0.0.1:5432}"
crate12_http = "[2017-02-01T15:43:37,793][INFO ][o.e.h.HttpServer         ] [8f64DTi] publish_address {127.0.0.1:4200}, bound_addresses {127.0.0.1:4200}"
crate12_transport = "[2017-02-01T15:43:37,783][INFO ][o.e.t.TransportService   ] [8f64DTi] publish_address {127.0.0.1:4300}, bound_addresses {127.0.0.1:4300}"


def get_match(lines):
    for line in lines:
        protocol, addr = AddrConsumer._parse(line)
        if protocol:
            return protocol, addr


class AddrParseTest(TestCase):

    def test_http_address(self):
        self.assertEqual(get_match(lines1), ('http', '10.68.2.10:4200'))
        self.assertEqual(get_match(lines2), ('http', '192.168.0.19:4200'))
        self.assertEqual(get_match(lines3), ('http', '127.0.0.1:42203'))
        self.assertEqual(get_match(lines4), ('http', '192.168.43.105:4200'))

    def test_crate12_psql(self):
        self.assertEqual(AddrConsumer._parse(crate12_psql), ('psql', '127.0.0.1:5432'))
        self.assertEqual(AddrConsumer._parse(crate12_http), ('http', '127.0.0.1:4200'))
        self.assertEqual(AddrConsumer._parse(crate12_transport), ('transport', '127.0.0.1:4300'))

    def test_ipv6_http(self):
        self.assertEqual(AddrConsumer._parse(ipv6_http), ('http', '[::1]:44200'))

    def test_ipv6_psql(self):
        self.assertEqual(AddrConsumer._parse(ipv6_psql), ('psql', '[::1]:5432'))


def load_tests(loader, tests, ignore):
    tests.addTests(DocTestSuite(run_crate))
    return tests


if __name__ == "__main__":
    main()



import argh
import os
import toml
import sys
from glob import glob
from .log import Logger
from .run_spec import do_run_spec
from .run_crate import CrateNode, get_crate
from .clients import client_errors


class Executor:
    def __init__(self,
                 *,
                 track_dir,
                 log,
                 sample_mode,
                 result_hosts=None,
                 crate_root=None,
                 fail_fast=None):
        self.track_dir = track_dir
        self.sample_mode = sample_mode
        self.result_hosts = result_hosts
        self.crate_root = crate_root
        self.log = log
        self.fail_fast = fail_fast

    def _expand_paths(self, paths):
        paths = (os.path.join(self.track_dir, path) for path in paths)
        paths = (os.path.abspath(path) for path in paths)
        return (p for path in paths for p in glob(path))

    def _run_specs(self, specs, benchmark_host, action=None):
        specs = self._expand_paths(specs)
        errors = []
        for spec in specs:
            self.log.info('### Running spec file: ', os.path.basename(spec))
            try:
                do_run_spec(
                    spec=spec,
                    benchmark_hosts=benchmark_host,
                    log=self.log,
                    result_hosts=self.result_hosts,
                    sample_mode=self.sample_mode,
                    action=action)
            except Exception:
                errors.append(True)
                if self.fail_fast:
                    raise
                else:
                    self.log.info('WARNING: Spec file failed due to the following exception:')
                    import traceback
                    traceback.print_exc()
        return any(errors)

    def _execute_specs(self, specs, benchmark_host):
        errors = []
        if isinstance(specs, list):
            errors.append(self._run_specs(specs, benchmark_host))
        else:
            fixtures = specs.get('fixtures', [])
            queries = specs.get('queries', [])
            full = specs.get('full', [])
            errors.append(
                self._run_specs(fixtures, benchmark_host, action=['setup']))
            errors.append(
                self._run_specs(queries, benchmark_host, action=['queries']))
            errors.append(
                self._run_specs(fixtures, benchmark_host, action=['teardown']))
            errors.append(
                self._run_specs(full, benchmark_host))
        return any(errors)

    def execute(self, track):
        configurations = list(self._expand_paths(track['configurations']))
        versions = track['versions']
        error = False
        for version in versions:
            self.log.info('# Version: ', version)
            for c, configuration in enumerate(configurations):
                self.log.info('## Starting Crate {0}, configuration: {1}'.format(
                    os.path.basename(version),
                    os.path.basename(configuration)
                ))
                configuration = toml.load(os.path.join(self.track_dir, configuration))
                crate_dir = get_crate(version, self.crate_root)
                with CrateNode(crate_dir=crate_dir,
                               env=configuration.get('env'),
                               settings=configuration.get('settings')) as node:
                    node.start()
                    _error = self._execute_specs(track['specs'], node.http_url)
                    error = error or _error
        return error


@argh.arg('-r', '--result_hosts', type=str)
@argh.arg('-of', '--output-fmt', choices=['json', 'text'], default='text')
@argh.arg('--failfast', action='store_true')
@argh.arg('--logfile-info', help='Redirect info messages to a file')
@argh.arg('--logfile-result', help='Redirect benchmark results to a file')
@argh.arg('--sample-mode', choices=('all', 'reservoir'),
          help='Method used for sampling', default='reservoir')
@argh.wrap_errors([KeyboardInterrupt, BrokenPipeError] + client_errors)
def run_track(track,
              result_hosts=None,
              crate_root=None,
              output_fmt=None,
              logfile_info=None,
              logfile_result=None,
              failfast=False,
              sample_mode='reservoir'):
    """Execute a track file"""
    with Logger(output_fmt=output_fmt,
                logfile_info=logfile_info,
                logfile_result=logfile_result) as log:
        executor = Executor(
            track_dir=os.path.dirname(track),
            log=log,
            result_hosts=result_hosts,
            crate_root=crate_root,
            fail_fast=failfast,
            sample_mode=sample_mode
        )
        error = executor.execute(toml.load(track))
        if error:
            sys.exit(1)


def main():
    argh.dispatch_command(run_track)


if __name__ == "__main__":
    main()



from cr8 import cli
from cr8.cli import dicts_from_stdin, lines_from_stdin, dicts_from_lines
from doctest import DocTestSuite
from unittest import TestCase, main
from unittest.mock import patch
import io


class CliTest(TestCase):

    @patch('sys.stdin',
           new_callable=lambda: io.StringIO('{"name": "n1"}\n{"name": "n2"}'))
    def test_dicts_from_stdin_multi_json(self, stdin):
        stdin.isatty = lambda: False

        dicts = iter(dicts_from_stdin())
        d1 = next(dicts)
        d2 = next(dicts)

        self.assertEqual({"name": "n1"}, d1)
        self.assertEqual({"name": "n2"}, d2)

        self.assertRaises(StopIteration, next, dicts)

    @patch('sys.stdin', new_callable=lambda: io.StringIO('{\n    "name": "n1"\n}'))
    def test_dicts_from_stdin_single_json(self, stdin):
        stdin.isatty = lambda: False

        dicts = iter(dicts_from_stdin())
        d1 = next(dicts)
        self.assertEqual({"name": "n1"}, d1)
        self.assertRaises(StopIteration, next, dicts)

    @patch('sys.stdin', new_callable=lambda: io.StringIO(
        '[\n{\n"name": "n1"\n},\n{\n"name": "n2"\n}\n]'))
    def test_dicts_from_json_list_of_obj(self, stdin):

        dicts = iter(dicts_from_stdin())
        d1 = next(dicts)
        d2 = next(dicts)

        self.assertEqual({"name": "n1"}, d1)
        self.assertEqual({"name": "n2"}, d2)

        self.assertRaises(StopIteration, next, dicts)

    @patch('sys.stdin', new_callable=io.StringIO)
    def test_lines_from_stdin_isatty_but_default(self, stdin):
        stdin.isatty = lambda: True

        lines = list(lines_from_stdin('default'))
        self.assertEqual(['default'], lines)

    def test_dicts_from_iterable(self):
        d = next(iter(dicts_from_lines(['{\n', '    "name": 10\n', '}\n'])))
        self.assertEqual(d, {"name": 10})

    def test_dicts_from_iterable_with_empty_lines(self):
        d = list(dicts_from_lines([
            '',
            '{"name": 10}',
            '\n',
            '',
            '{"name": 20}\n'
            '\n'
        ]))
        self.assertEqual(d[0], {"name": 10})
        self.assertEqual(d[1], {"name": 20})

    def test_dicts_from_iterator(self):
        d = next(iter(dicts_from_lines(iter(['{\n', '    "name": 10\n', '}\n']))))
        self.assertEqual(d, {"name": 10})


def load_tests(loader, tests, ignore):
    tests.addTests(DocTestSuite(cli))
    return tests


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# PYTHON_ARGCOMPLETE_OK

import sys
import argh
import argparse
import logging
from subprocess import run, CalledProcessError

from cr8 import __version__
from cr8.misc import break_iterable, init_logging
from cr8.timeit import timeit
from cr8.insert_json import insert_json
from cr8.insert_fake_data import insert_fake_data
from cr8.insert_blob import insert_blob
from cr8.run_spec import run_spec
from cr8.run_crate import run_crate, create_node
from cr8.run_track import run_track


log = logging.getLogger(__name__)


def _run_subcommand(parser, args):
    if args[0][0] == '@':
        args[0] = args[0][1:]
        try:
            run(args, check=True)
        except CalledProcessError:
            sys.exit('Failure running: ' + ' '.join(args))

    else:
        parser.dispatch(args)


def _run_crate_and_rest(parser, args_groups):
    args = parser.parse_args(args_groups[0])
    log.info('# run-crate')
    log.info('===========\n')
    with create_node(version=args.version,
                     env=args.env,
                     setting=args.setting,
                     crate_root=args.crate_root,
                     keep_data=args.keep_data) as node:
        node.start()
        log.info('\n')
        for args in args_groups[1:]:
            if not args or not args[0]:
                continue
            cmd = '# ' + args[0]
            log.info(cmd)
            log.info('=' * len(cmd) + '\n')
            for i in range(len(args)):
                try:
                    args[i] = args[i].format(node=node)
                except KeyError:
                    pass
            _run_subcommand(parser, args)
            log.info('\n')


def main():
    p = argh.ArghParser(
        prog='cr8', formatter_class=argparse.RawTextHelpFormatter)
    p.add_argument(
        '--version', action='version', version="%(prog)s " + __version__)
    p.add_commands([timeit,
                    insert_json,
                    insert_fake_data,
                    insert_blob,
                    run_spec,
                    run_crate,
                    run_track])
    args_groups = list(break_iterable(sys.argv[1:], lambda x: x == '--'))
    if len(args_groups) == 1:
        p.dispatch()
        return

    init_logging(log)
    if args_groups[0][0] == 'run-crate':
        _run_crate_and_rest(p, args_groups)
    else:
        for args in args_groups:
            p.dispatch(args)


if __name__ == '__main__':
    main()




import sys
import json
import ast
from collections import OrderedDict


def to_int(s):
    """ converts a string to an integer

    >>> to_int('1_000_000')
    1000000

    >>> to_int('1e6')
    1000000

    >>> to_int('1000')
    1000
    """
    try:
        return int(s.replace('_', ''))
    except ValueError:
        return int(ast.literal_eval(s))


def lines_from_stdin(default=None):
    if sys.stdin.isatty():
        if default:
            yield default
        return
    for line in sys.stdin:
        yield line


def dicts_from_lines(lines):
    """ returns a generator producing dicts from json lines

    1 JSON object per line is supported:

        {"name": "n1"}
        {"name": "n2"}

    Or 1 JSON object:

        {
            "name": "n1"
        }

    Or a list of JSON objects:

        [
            {"name": "n1"},
            {"name": "n2"},
        ]
    """
    lines = iter(lines)
    for line in lines:
        line = line.strip()
        if not line:
            continue  # skip empty lines
        try:
            yield json.loads(line, object_pairs_hook=OrderedDict)
        except json.decoder.JSONDecodeError:
            content = line + ''.join(lines)
            dicts = json.loads(content, object_pairs_hook=OrderedDict)
            if isinstance(dicts, list):
                yield from dicts
            else:
                yield dicts


def dicts_from_stdin():
    if sys.stdin.isatty():
        raise SystemExit('Expected json input via stdin')
    yield from dicts_from_lines(sys.stdin)


dicts_from_stdin.__doc__ = dicts_from_lines.__doc__




from unittest import TestCase
from doctest import DocTestSuite
from cr8.engine import eval_fail_if, Result, FailIf, TimedStats
from cr8.metrics import Stats
from cr8 import engine


class FailIfTest(TestCase):

    def test_fail_if_fails(self):
        timed_stats = TimedStats(1, 2, Stats())
        result = Result({}, 'select name', timed_stats, 1)
        with self.assertRaises(FailIf):
            eval_fail_if("'{statement}' == 'select name'", result)

    def test_fail_if_supports_runtime_stats(self):
        stats = Stats()
        stats.measure(103.2)
        stats.measure(205.2)
        timed_stats = TimedStats(1, 2, stats)
        result = Result({}, 'select name', timed_stats, 1)
        with self.assertRaises(FailIf):
            eval_fail_if("{runtime_stats.max} > 30", result)

    def test_fail_if_supports_concurrency(self):
        stats = Stats()
        timed_stats = TimedStats(1, 2, stats)
        result = Result({}, 'select name', timed_stats, 1)
        with self.assertRaises(FailIf):
            eval_fail_if("{concurrency} == 1", result)

    def test_fail_if_supports_bulk_size(self):
        stats = Stats()
        timed_stats = TimedStats(1, 2, stats)
        result = Result({}, 'select name', timed_stats, 1, bulk_size=200)
        eval_fail_if("{bulk_size} < 200", result)


def load_tests(loader, tests, ignore):
    tests.addTests(DocTestSuite(engine))
    return tests



from unittest import main, TestCase
from doctest import DocTestSuite

from cr8 import metrics


class UniformReservoirTest(TestCase):

    def test_fewer_values_than_size(self):
        r = metrics.UniformReservoir(10)
        r.add(10)
        r.add(20)
        self.assertEqual([10, 20], r.values)


class StatsTest(TestCase):

    def test_stats_are_empty_without_values(self):
        hist = metrics.Stats(metrics.All)
        result = hist.get()
        self.assertEqual(len(result), 1)
        self.assertEqual(result['n'], 0)

    def test_stats_has_min_max_mean_samples_and_stdev_with_1_value(self):
        hist = metrics.Stats()
        hist.measure(23.2)
        result = hist.get()
        self.assertEqual(result['mean'], 23.2)
        self.assertEqual(result['min'], 23.2)
        self.assertEqual(result['max'], 23.2)
        self.assertEqual(result['samples'], [23.2])
        self.assertEqual(result['stdev'], 0)
        self.assertEqual(result['n'], 1)

    def test_stats(self):
        hist = metrics.Stats()
        hist.measure(10.5)
        hist.measure(38.1)
        hist.measure(234.7)
        hist.measure(50.2)

        result = hist.get()
        self.assertEqual(result['min'], 10.5)
        self.assertEqual(result['max'], 234.7)
        self.assertEqual(result['mean'], 83.375)
        self.assertEqual(round(result['median'], 3), 44.150)
        self.assertEqual(round(result['variance'], 3), 10453.476)
        self.assertEqual(round(result['stdev'], 3), 102.242)

        self.assertEqual(result['percentile']['50'], 38.1)
        self.assertEqual(result['percentile']['75'], 50.2)
        self.assertEqual(result['percentile']['99'], 234.7)

    def test_n_is_number_of_iterations(self):
        hist = metrics.Stats()
        for i in range(10):
            hist.measure(i)
        self.assertEqual(hist.get()['n'], 10)


def load_tests(loader, tests, ignore):
    tests.addTests(DocTestSuite(metrics))
    return tests


if __name__ == "__main__":
    main()



import itertools
from functools import partial
from time import time, perf_counter
from collections import namedtuple

from cr8 import aio
from cr8.metrics import Stats, get_sampler
from cr8.clients import client


TimedStats = namedtuple('TimedStats', ['started', 'ended', 'stats'])


class FailIf(SystemExit):
    pass


def eval_fail_if(fail_if: str, result):
    fail_if = fail_if.format(**result.as_dict())
    if eval(fail_if):
        raise FailIf("Expression failed: " + fail_if)


class DotDict(dict):

    __getattr__ = dict.__getitem__


class Result:
    def __init__(self,
                 version_info,
                 statement,
                 timed_stats,
                 concurrency,
                 meta=None,
                 bulk_size=None):
        self.version_info = version_info
        self.statement = str(statement)
        self.meta = meta and DotDict(meta) or None
        self.started = timed_stats.started
        self.ended = timed_stats.ended
        self.runtime_stats = DotDict(timed_stats.stats.get())
        self.concurrency = concurrency
        self.bulk_size = bulk_size

    def as_dict(self):
        return self.__dict__


def run_and_measure(f, statements, concurrency, num_items=None, sampler=None):
    stats = Stats(sampler)
    measure = partial(aio.measure, stats, f)
    started = int(time() * 1000)
    aio.run_many(measure, statements, concurrency, num_items=num_items)
    ended = int(time() * 1000)
    return TimedStats(started, ended, stats)


def _generate_statements(stmt, args, iterations, duration):
    if duration is None:
        yield from itertools.repeat((stmt, args), iterations or 100)
    else:
        now = perf_counter()
        while perf_counter() - now < duration:
            yield (stmt, args)


class Runner:
    def __init__(self, hosts, concurrency, sample_mode):
        self.concurrency = concurrency
        self.client = client(hosts, concurrency=concurrency)
        self.sampler = get_sampler(sample_mode)

    def warmup(self, stmt, num_warmup):
        statements = itertools.repeat((stmt,), num_warmup)
        aio.run_many(self.client.execute, statements, 0, num_items=num_warmup)

    def run(self, stmt, *, iterations=None, duration=None, args=None, bulk_args=None):
        if bulk_args:
            args = bulk_args
            f = self.client.execute_many
        else:
            f = self.client.execute
        statements = _generate_statements(stmt, args, iterations, duration)
        return run_and_measure(
            f, statements, self.concurrency, iterations, sampler=self.sampler)

    def __enter__(self):
        return self

    def __exit__(self, *ex):
        self.client.close()



#!/usr/bin/env python

from setuptools import setup

try:
    with open('README.rst', 'r', encoding='utf-8') as f:
        readme = f.read()
except IOError:
    readme = ''


setup(
    name='cr8',
    author='Mathias Fußenegger',
    author_email='pip@zignar.net',
    url='https://github.com/mfussenegger/cr8',
    description='A collection of command line tools for crate devs',
    long_description=readme,
    entry_points={
        'console_scripts': [
            'cr8 = cr8.__main__:main',
        ]
    },
    packages=['cr8'],
    install_requires=[
        'crate>=0.16',
        'argh',
        'tqdm',
        'Faker==0.9.1',
        'aiohttp>=3.3,<4',
        'toml'
    ],
    extras_require={
        'uvloop': ['uvloop'],
        'asyncpg': ['asyncpg']
    },
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    use_scm_version=True,
    setup_requires=['setuptools_scm']
)



from cr8.insert_fake_data import DataFaker
from cr8 import insert_fake_data
from unittest import TestCase, main
from doctest import DocTestSuite
from decimal import Decimal
import datetime


class TestDataFaker(TestCase):

    def setUp(self):
        self.f = DataFaker()
        self.f.fake.seed_instance(42)

    def test_fake_provider_for_name_column(self):
        provider = self.f.provider_for_column('name', 'string')
        self.assertEqual(provider(), 'Allison Hill')

    def test_fake_provider_for_string_id_column(self):
        provider = self.f.provider_for_column('id', 'string')
        # even with seed set the uuid is not deterministic.. just check the length
        self.assertEqual(len(provider()), len('8731cdac-8671-441d-b07f-e766ffe303e1'))

    def test_fake_provider_for_int_id_column(self):
        provider = self.f.provider_for_column('id', 'integer')
        self.assertEqual(provider(), 1)
        self.assertEqual(provider(), 2)
        provider2 = self.f.provider_for_column('id', 'integer')
        self.assertEqual(provider2(), 1)

    def test_custom_auto_inc(self):
        provider = self.f.provider_for_column('auto_inc', 'long')
        self.assertEqual(provider(), 1)
        self.assertEqual(provider(), 2)

    def test_fake_provider_for_long_id_column(self):
        provider = self.f.provider_for_column('id', 'long')
        self.assertEqual(provider(), 1)

    def test_type_default_provider_for_unknown_int_column(self):
        provider = self.f.provider_for_column(
            'column_name_without_provider', 'integer')
        self.assertEqual(provider(), 598833565)  # got random_int provider

    def test_timestamp_column_default(self):
        provider = self.f.provider_for_column('timestamp', 'timestamp')
        dt = provider()
        diff = datetime.datetime(2017, 11, 18, 19, 0, 0) - dt
        self.assertLessEqual(diff, datetime.timedelta(seconds=1))

    def test_timestamp_type_default(self):
        provider = self.f.provider_for_column('some_ts_column', 'timestamp')
        dt = provider()
        diff = datetime.datetime(2017, 11, 18, 19, 0, 0) - dt
        self.assertLessEqual(diff, datetime.timedelta(seconds=1))

    def test_provider_from_mapping(self):
        mapping = {'x': ['random_int', [10, 20]]}
        provider = self.f.provider_from_mapping('x', mapping)
        self.assertEqual(provider(), 20)

    def test_float_type_default(self):
        provider = self.f.provider_for_column('x', 'float')
        self.assertEqual(provider(), -31246013015.0)

    def test_double_type_default(self):
        provider = self.f.provider_for_column('x', 'double')
        self.assertEqual(provider(), Decimal(-31246013015.0))

    def test_short_type_default(self):
        provider = self.f.provider_for_column('x', 'short')
        self.assertEqual(provider(), -18176)

    def test_byte_type_default(self):
        provider = self.f.provider_for_column('x', 'byte')
        self.assertEqual(provider(), -71)

    def test_ip_type_default(self):
        provider = self.f.provider_for_column('x', 'ip')
        self.assertEqual(provider(), '198.50.12.206')

    def test_geopoint_type_default(self):
        provider = self.f.provider_for_column('location', 'geo_point')
        self.assertEqual(provider(), [50.19364744483815, -85.49806405991995])

    def test_geoshape_type_default(self):
        provider = self.f.provider_for_column('area', 'geo_shape')
        self.assertEqual(provider(),
                         'POLYGON (( '
                         '-132.47048275355667 44.147296981090086, '
                         '-131.2679223792111 42.09369742374501, '
                         '-132.14018682738413 37.17894586552094, '
                         '-133.04540290479406 36.10457754779138, '
                         '-142.31051949147854 46.75961787621673, '
                         '-132.47048275355667 44.147296981090086 '
                         '))')

    def test_invalid_provider_for_column(self):
        msg = 'No fake provider found for column "x" with type "y"'
        with self.assertRaises(ValueError) as cm:
            self.f.provider_for_column('x', 'y')
        self.assertEqual(str(cm.exception), msg)

    def test_provider_for_object_column_creates_empty_dicts(self):
        provider = self.f.provider_for_column('obj', 'object')
        self.assertEqual(provider(), dict())

    def test_provider_for_string_array(self):
        provider = self.f.provider_for_column('foo', 'string_array')
        value = provider()
        self.assertEqual(len(value), 40)
        self.assertEqual(value[0], 'born')

    def test_provider_for_nested_string_array(self):
        provider = self.f.provider_for_column('foo', 'string_array_array')
        value = provider()
        self.assertEqual(len(value), 40)
        self.assertEqual(len(value[0]), 7)
        self.assertEqual(value[0][0:2], ['agent', 'every'])


def load_tests(loader, tests, ignore):
    tests.addTests(DocTestSuite(insert_fake_data))
    return tests


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-


import argh
from crate.client import connect


@argh.arg('filename', help='path/filename of the file that should be uploaded')
@argh.arg('--table', help='name of the blob table', required=True)
@argh.arg('--hosts', type=str,
          help='crate hosts to which the file should be uploaded to',
          default=['http://localhost:4200'])
@argh.wrap_errors([KeyboardInterrupt])
def insert_blob(filename, hosts=None, table=None):
    """Upload a file into a blob table """
    conn = connect(hosts)
    container = conn.get_blob_container(table)
    with open(filename, 'rb') as f:
        digest = container.put(f)
    return '{server}/_blobs/{table}/{digest}'.format(
        server=conn.client.active_servers[0],
        table=table,
        digest=digest
    )


def main():
    argh.dispatch_command(insert_blob)


if __name__ == '__main__':
    main()



import doctest
from unittest import main
from cr8 import fake_providers


def load_tests(loader, tests, ignore):
    flags = (doctest.ELLIPSIS | doctest.IGNORE_EXCEPTION_DETAIL)
    tests.addTests(doctest.DocTestSuite(fake_providers, optionflags=flags))
    return tests


if __name__ == "__main__":
    main()
